"""
此处定义了一些 Pydantic 模型，使用 Pydantic V2
"""
from enum import IntEnum
from typing import Optional, Any, Union

from pydantic import BaseModel, UUID4, ConfigDict, field_serializer, AliasGenerator, model_validator, \
    field_validator
from pydantic.alias_generators import to_camel, to_snake
from pydantic_core.core_schema import SerializerFunctionWrapHandler, FieldSerializationInfo

from .enums import MessageType, RetCode, MessageDataHead

__all__ = ("WS_MESSAGE_MAX_LENGTH", "WebSocketMessage", "StrengthData")

WS_MESSAGE_MAX_LENGTH = 1950
"""WebSocket 消息最大长度"""


# noinspection PyNestedDecorators
class WebSocketMessage(BaseModel):
    """
    WebSocket 消息

    :ivar type: 消息类型
    :ivar client_id: 第三方终端 ID
    :ivar target_id: App ID
    :ivar message: 消息 / 指令
    """
    # 仅在序列化时将键转为驼峰命名法作为别名
    # 验证时不取别名的原因是，取别名会影响对象初始化参数名，而 AliasChoices 不知为何不起作用
    model_config = ConfigDict(
        alias_generator=AliasGenerator(
            serialization_alias=to_camel
        )
    )

    type: MessageType
    client_id: Optional[UUID4] = None
    target_id: Optional[UUID4] = None
    message: Union[str, RetCode, MessageDataHead]

    @field_validator("message")
    @classmethod
    def _validate_message(cls, value: Any):
        """自动先行尝试解析 `message`"""
        if isinstance(value, str):
            try:
                return RetCode(int(value))
            except ValueError:
                try:
                    return MessageDataHead(value)
                except ValueError:
                    pass
        return value

    @field_serializer("message", mode="wrap")
    def _serialize_message(self, value: Any, nxt: SerializerFunctionWrapHandler, _: FieldSerializationInfo):
        """对于 ``IntEnum`` 的枚举，转化为 ``str``"""
        if isinstance(value, IntEnum):
            return str(value.value)
        else:
            return nxt(value)

    @field_serializer("client_id", "target_id")
    def _serialize_id(self, value: Optional[UUID4]):
        """序列化时，当值为 ``None``，序列化成空字符串"""
        return value or ""

    @field_validator("client_id", "target_id", mode="before")
    @classmethod
    def _validate_id(cls, value: Any):
        """验证 UUID 值时，当值为空字符串，更改为 ``None``，与 :meth:`serialize_id` 相对应"""
        return value if value != "" else None

    @model_validator(mode="before")
    @classmethod
    def _field_to_pascal(cls, data: Any):
        """
        验证模型时，将所有的键名更改为蛇形命名法（下划线）

        验证时不取别名的原因是，取别名会影响对象初始化参数名，而 ``AliasChoices`` 不知为何不起作用
        """
        if isinstance(data, dict):
            for key in list(data.keys()):
                data[to_snake(key)] = data.pop(key)
        return data


class StrengthData(BaseModel):
    """
    强度数据

    :ivar a: A 通道强度
    :ivar b: B 通道强度
    :ivar a_limit: A 通道强度上限
    :ivar b_limit: B 通道强度上限
    """
    a: int
    b: int
    a_limit: int
    b_limit: int



"""
此处定义了一些异常类
"""
__all__ = ("InvalidStrengthData", "InvalidFeedbackData", "InvalidPulseOperation", "PulseDataTooLong")

from typing import Any


class InvalidStrengthData(Exception):
    """强度数据不合法"""

    def __init__(self, strength_data: str):
        super().__init__(f"Invalid strength data: {strength_data}")


class InvalidFeedbackData(Exception):
    """App 反馈按钮数据不合法"""

    def __init__(self, feedback_data: str):
        super().__init__(f"Invalid strength data: {feedback_data}")


class InvalidPulseOperation(Exception):
    """波形操作数据不合法"""

    def __init__(self, pulse_operation: Any):
        super().__init__(f"Invalid pulse operation: {pulse_operation}")


class PulseDataTooLong(Exception):
    """波形操作数据列表过长"""

    def __init__(self, length: int):
        super().__init__(f"Pulse data too long: {length}")



from abc import ABC, abstractmethod
from typing import Optional, AsyncGenerator, Any, Union, Dict, Callable, Literal, Type, TypeVar

from pydantic import UUID4

from ..enums import MessageDataHead, RetCode, StrengthOperationType, Channel, FeedbackButton, MessageType
from ..models import StrengthData
from ..models import WebSocketMessage
from ..typing import PulseOperation
from ..utils import dg_lab_client_qrcode, parse_strength_data, parse_feedback_data, dump_strength_operation, \
    dump_add_pulses, dump_clear_pulses

__all__ = ["DGLabClient"]

_DataType = TypeVar("_DataType", Type[StrengthData], Type[FeedbackButton], Type[RetCode])


class DGLabClient(ABC):
    """
    DG-Lab 终端基础类

    :param client_id: 终端 ID
    :param target_id: App ID
    """

    def __init__(
            self,
            client_id: UUID4 = None,
            target_id: UUID4 = None
    ):
        self._client_id: Optional[UUID4] = client_id
        self._target_id: Optional[UUID4] = target_id
        self._message_type_to_handler: Dict[
            MessageType,
            Callable[[WebSocketMessage], Any]
        ] = {
            MessageType.MSG: self._handle_msg,
            MessageType.BREAK: self._handle_break,
            MessageType.HEARTBEAT: self._handle_heartbeat
        }

    @property
    def client_id(self) -> Optional[UUID4]:
        """DG-Lab 终端 ID"""
        return self._client_id

    @property
    def target_id(self) -> Optional[UUID4]:
        """DG-Lab App ID"""
        return self._target_id

    @property
    def not_registered(self) -> bool:
        """终端是否未注册"""
        return self._client_id is None

    @property
    def not_bind(self) -> bool:
        """终端是否未完成与 App 的绑定"""
        return self._client_id is None or self.target_id is None

    @abstractmethod
    async def _recv(self) -> WebSocketMessage:
        """
        收取来自 WebSocket 服务端的消息，并解析为 [`WebSocketMessage`][pydglab_ws.models.WebSocketMessage]
        """
        ...

    @abstractmethod
    async def _send(self, message: WebSocketMessage):
        """
        向 WebSocket 服务端发送消息

        :param message: 解析为 [`WebSocketMessage`][pydglab_ws.models.WebSocketMessage] 的消息
        """
        ...

    def get_qrcode(self, uri: str) -> Optional[str]:
        """
        终端二维码，二维码图像需要自行生成

        :param uri: WebSocket 服务端 URI，例如：``ws://107.47.91.92:4567``
            （注意末尾不能有 ``/``）
        :return: URL 字符串，如果需要二维码图像需要自行从返回的文本进行生成
        """
        if uri is None or self.not_registered:
            return None
        return dg_lab_client_qrcode(uri, self._client_id)

    async def _recv_owned(self) -> WebSocketMessage:
        """
        与 :meth:`_recv` 类似，但只接收目标为自身终端的消息
        """
        message = await self._recv()
        if message.client_id == self._client_id:
            return message

    async def _send_owned(self, msg_type: MessageType, msg: str):
        """
        与 :meth:`_send` 类似，但代为设置 ``client_id``, ``target_id``

        :param msg_type: :attr:`WebSocketMessage.type`
        :param msg: :attr:`WebSocketMessage.message`
        """
        message = WebSocketMessage(
            type=msg_type,
            client_id=self._client_id,
            target_id=self._target_id,
            message=msg
        )
        await self._send(message)

    @staticmethod
    def _handle_msg(message: WebSocketMessage) -> Optional[Union[StrengthData, FeedbackButton, RetCode]]:
        """
        处理类型为 ``msg`` 的消息

        :raise InvalidStrengthData: [`InvalidStrengthData`][pydglab_ws.exceptions.InvalidStrengthData]
        :raise InvalidFeedbackData: [`InvalidFeedbackData`][pydglab_ws.exceptions.InvalidFeedbackData]
        """
        if isinstance(message.message, str):
            if message.message.startswith(MessageDataHead.STRENGTH.value):
                return parse_strength_data(message.message)
            elif message.message.startswith(MessageDataHead.FEEDBACK.value):
                return parse_feedback_data(message.message)
        elif isinstance(message.message, RetCode):
            return message.message
        return None

    @staticmethod
    def _handle_break(message: WebSocketMessage) -> Optional[Literal[RetCode.CLIENT_DISCONNECTED]]:
        """处理类型为 ``break`` 的消息"""
        return message.message

    @staticmethod
    def _handle_heartbeat(message: WebSocketMessage) -> Optional[Literal[RetCode.SUCCESS]]:
        """处理类型为 ``heartbeat`` 的消息"""
        return message.message

    async def register(self):
        """
        从 WebSocket 服务端中获取 ``client_id`` 并保存
        """
        while self.not_registered:
            message = await self._recv()
            if message.type == MessageType.BIND and message.message == MessageDataHead.TARGET_ID:
                self._client_id = message.client_id

    async def ensure_bind(self):
        """确保终端已完成与 App 的绑定"""
        while True:
            if self.not_registered:
                await self.register()
            elif self.not_bind:
                await self.bind()
            else:
                break

    async def bind(self) -> RetCode:
        """
        等待与 DG-Lab App 的关系绑定，并保存 ``target_id``
        :return: 响应码
        """
        while self.not_bind:
            message = await self._recv_owned()
            if message.type == MessageType.BIND and isinstance(message.message, RetCode):
                if message.message == RetCode.SUCCESS:
                    self._target_id = message.target_id
                return message.message

    async def rebind(self) -> RetCode:
        """
        清除 ``target_id``，重新等待与 DG-Lab App 的关系绑定，适合 App 断开连接后调用
        :return: 响应码
        """
        self._target_id = None
        while self.not_bind:
            message = await self._recv_owned()
            if message.type == MessageType.BIND and isinstance(message.message, RetCode):
                if message.message == RetCode.SUCCESS:
                    self._target_id = message.target_id
                return message.message

    async def recv_data(self) -> Union[StrengthData, FeedbackButton, RetCode]:
        """
        获取 WebSocket 服务端的数据

        注意，获取到的是队列中最早的数据，可能不是最新的

        :return: 可能为 **强度数据** - [`StrengthData`][pydglab_ws.models.StrengthData]、 \
            **App 反馈数据** - [`FeedbackButton`][pydglab_ws.enums.FeedbackButton]、 \
            **心跳** - [`RetCode.SUCCESS`][pydglab_ws.enums.RetCode.SUCCESS]、 \
            **App 断开连接** - [`RetCode.CLIENT_DISCONNECTED`][pydglab_ws.enums.RetCode.CLIENT_DISCONNECTED]、\
            **消息长度大于 1950** - [`RetCode.MESSAGE_TOO_LONG`][pydglab_ws.enums.RetCode.MESSAGE_TOO_LONG]
        :raise InvalidStrengthData: [`InvalidStrengthData`][pydglab_ws.exceptions.InvalidStrengthData]
        :raise InvalidFeedbackData: [`InvalidFeedbackData`][pydglab_ws.exceptions.InvalidFeedbackData]
        """
        await self.ensure_bind()
        while True:
            message = await self._recv_owned()
            handler = self._message_type_to_handler.get(message.type)
            if handler and (result := handler(message)) is not None:
                return result

    async def data_generator(
            self,
            *targets: _DataType,
    ) -> AsyncGenerator[_DataType, Any]:
        """
        强度数据异步生成器

        注意，是从队列中最早的数据开始获取，可能不是最新的

        示例：
        ```python3
        async for data in client.data_generator(StrengthData, FeedbackButton):
            print(f"Got data from App: {data}")
        ```
        :param targets: 目标类型，只有为目标类型的数据会被返回，为空即默认值时则不进行限制
        :return: 可能为 **强度数据** - [`StrengthData`][pydglab_ws.models.StrengthData]、 \
            **App 反馈数据** - [`FeedbackButton`][pydglab_ws.enums.FeedbackButton] \
            、**心跳** - ``RetCode.SUCCESS``、**App 断开连接** - ``RetCode.CLIENT_DISCONNECTED``
        """
        while True:
            data = await self.recv_data()
            if not targets or type(data) in targets:
                yield data

    async def set_strength(
            self,
            channel: Channel,
            operation_type: StrengthOperationType,
            value: int
    ):
        """
        设置强度

        :param channel: 通道选择
        :param operation_type: 强度变化模式
        :param value: 强度数值，范围在 [0, 200]
        """
        await self.ensure_bind()
        await self._send_owned(
            MessageType.MSG,
            dump_strength_operation(channel, operation_type, value)
        )

    async def add_pulses(
            self,
            channel: Channel,
            *pulses: PulseOperation
    ):
        """
        下发波形数据

        - 每条波形数据代表了 100ms 的数据，所以若每次发送的数据有 10 条，那么就是 1s 的数据。 \
          由于网络有一定延时，若要保证波形输出的连续性，建议波形数据的发送间隔略微小于波形数据的时间长度 (< 1s)
        - 数组最大长度为 100,也就是最多放置 10s 的数据，另外 App 中的波形队列最大长度为 500，即为 50s 的数据， \
          若后接收到的数据无法全部放入波形队列，多余的部分会丢弃。所以谨慎考虑您的数据长度和数据发送间隔

        :param channel: 通道选择
        :param pulses: 波形操作数据，最大长度为 100
        :raise InvalidPulseOperation: [`InvalidPulseOperation`][pydglab_ws.exceptions.InvalidPulseOperation]
        :raise PulseDataTooLong: 波形操作数据过长，最大长度应为 [`PULSE_DATA_MAX_LENGTH`][pydglab_ws.utils.PULSE_DATA_MAX_LENGTH]
        """
        await self.ensure_bind()
        await self._send_owned(
            MessageType.MSG,
            dump_add_pulses(channel, *pulses)
        )

    async def clear_pulses(self, channel: Channel):
        """
        清空波形队列

        :param channel: 通道选择
        :raise InvalidPulseOperation: [`InvalidPulseOperation`][pydglab_ws.exceptions.InvalidPulseOperation]
        """
        await self.ensure_bind()
        await self._send_owned(
            MessageType.MSG,
            dump_clear_pulses(channel)
        )



"""
此处定义了一些常量枚举
"""
import enum
from enum import Enum, IntEnum

__all__ = (
    "MessageType",
    "RetCode",
    "MessageDataHead",
    "StrengthOperationType",
    "FeedbackButton",
    "Channel"
)


@enum.unique
class MessageType(str, Enum):
    """
    WebSocket 消息类型

    :ivar HEARTBEAT: 心跳包数据
    :ivar BIND: ID 关系绑定
    :ivar MSG: 波形下发/强度变化/队列清空等数据指令
    :ivar BREAK: 连接断开
    :ivar ERROR: 服务错误
    """
    HEARTBEAT = "heartbeat"
    BIND = "bind"
    MSG = "msg"
    BREAK = "break"
    ERROR = "error"


@enum.unique
class RetCode(IntEnum):
    """
    WebSocket 消息错误码枚举

    :ivar SUCCESS: 成功
    :ivar CLIENT_DISCONNECTED: 对方客户端已断开
    :ivar INVALID_CLIENT_ID: 二维码中没有有效的 ``clientId``
    :ivar SERVER_DELAY: Socket 连接上了，但服务器迟迟不下发 App 端的 ID 来绑定
    :ivar ID_ALREADY_BOUND: 此 ID 已被其他客户端绑定关系
    :ivar TARGET_CLIENT_NOT_FOUND: 要绑定的目标客户端不存在
    :ivar INCOMPATIBLE_RELATIONSHIP: 收信方和寄信方不是绑定关系
    :ivar NON_JSON_CONTENT: 发送的内容不是标准 JSON 对象
    :ivar RECIPIENT_NOT_FOUND: 未找到收信人（离线）
    :ivar MESSAGE_TOO_LONG: 下发的 ``message`` 长度大于 1950
    :ivar SERVER_INTERNAL_ERROR: 服务器内部异常
    """
    SUCCESS = 200
    CLIENT_DISCONNECTED = 209
    INVALID_CLIENT_ID = 210
    SERVER_DELAY = 211
    ID_ALREADY_BOUND = 400
    TARGET_CLIENT_NOT_FOUND = 401
    INCOMPATIBLE_RELATIONSHIP = 402
    NON_JSON_CONTENT = 403
    RECIPIENT_NOT_FOUND = 404
    MESSAGE_TOO_LONG = 405
    SERVER_INTERNAL_ERROR = 500


@enum.unique
class MessageDataHead(str, Enum):
    """
    WebSocket 消息数据开头部分

    :ivar TARGET_ID: Socket 通道与终端绑定
    :ivar DG_LAB: 关系绑定
    :ivar STRENGTH: 强度操作
    :ivar PULSE: 波形操作
    :ivar CLEAR: 清空波形队列
    :ivar FEEDBACK: App 反馈
    """
    TARGET_ID = "targetId"
    # noinspection SpellCheckingInspection
    DG_LAB = "DGLAB"
    STRENGTH = "strength"
    PULSE = "pulse"
    CLEAR = "clear"
    FEEDBACK = "feedback"


@enum.unique
class StrengthOperationType(IntEnum):
    """
    强度变化模式

    :ivar DECREASE: 通道强度减少
    :ivar INCREASE: 通道强度增加
    :ivar SET_TO: 通道强度变化为指定数值
    """
    DECREASE = 0
    INCREASE = 1
    SET_TO = 2


# class StrengthOperationType(Enum):
#     """
#     强度变化模式
#
#     :ivar NONE: 代表对应通道强度不做改变，对应通道的强度设定值无论是什么都无效
#     :ivar INCREASE: 代表对应通道强度相对增加变化
#     :ivar DECREASE: 代表对应通道强度相对减少变化
#     :ivar SET_TO: 代表对应通道强度绝对变化
#     """
#     NONE = 0b00
#     INCREASE = 0b01
#     DECREASE = 0b10
#     SET_TO = 0b11


@enum.unique
class FeedbackButton(IntEnum):
    """
    App 反馈按钮

    * A 通道 5 个按钮（从左至右）的角标为 0,1,2,3,4
    * B 通道 5 个按钮（从左至右）的角标为 5,6,7,8,9
    """
    A1 = 0
    A2 = 1
    A3 = 2
    A4 = 3
    A5 = 4
    B1 = 5
    B2 = 6
    B3 = 7
    B4 = 8
    B5 = 9


@enum.unique
class Channel(IntEnum):
    """
    通道

    :ivar A: A 通道
    :ivar B: B 通道
    """
    A = 1
    B = 2



import asyncio
from typing import Callable, Any, Coroutine

from pydantic import UUID4

from .base import DGLabClient
from ..models import WebSocketMessage

__all__ = ["DGLabLocalClient"]

try:
    WebSocketMessageQueue = asyncio.Queue[WebSocketMessage]
except TypeError:
    WebSocketMessageQueue = asyncio.Queue


class DGLabLocalClient(DGLabClient):
    # noinspection SpellCheckingInspection
    """
        DG-Lab 终端，并不实际发送 WebSocket 消息，而是直接与本地服务端进行通信

        本地服务端指同一线程下的 [`DGLabWSServer`][pydglab_ws.server.server.DGLabWSServer]

        :param client_id: 终端 ID
        :param sender: 用于客户端发送消息的回调函数
        :param queue_setter: 回调函数，用于服务端设置客户端的消息队列
        :param max_queue: 消息队列最大长度
        """

    def __init__(
            self,
            client_id: UUID4,
            sender: Callable[[WebSocketMessage], Coroutine[Any, Any, Any]],
            queue_setter: Callable[[UUID4, WebSocketMessageQueue], Any],
            max_queue: int = 2 ** 5
    ):
        super().__init__()
        self._client_id = client_id
        self._send_callable = sender
        self._message_queue: WebSocketMessageQueue = asyncio.Queue(max_queue)
        queue_setter(client_id, self._message_queue)

    async def _recv(self) -> WebSocketMessage:
        return await self._message_queue.get()

    async def _send(self, message: WebSocketMessage):
        await self._send_callable(message)
